module Evaluation.Function
  ( evalFunction
  ) where

import Prelude hiding (String, FilePath)
import qualified Prelude

import Types.Values
import Types (Function(..), Value(..), fromValue, makeString, makeImage)

import qualified Codec.Picture as CP
import qualified Codec.Picture.STBIR as CPS
import Data.Char (toLower)
import qualified Data.Text as T
import qualified Data.Text.Lazy as TL
import qualified Text.Pandoc as P
import qualified Text.Blaze.Html.Renderer.Text as B

fileComponents :: Prelude.String -> (Prelude.String, Prelude.String)
fileComponents s =
  let (lastRev, firstRev) = span (/= '.') $ reverse s
  in case firstRev of
       _ : firstRev' -> (reverse firstRev', reverse lastRev)
       [] -> (reverse lastRev, "")

unStringWrapper :: Value -> Prelude.String
unStringWrapper = \case
  String (StringWrapper s) -> s
  _ -> error "unexpected"

evalFunction :: Function -> Value -> Value
evalFunction f x = case (f, x) of
  (AppendStrings, Tuple (String (StringWrapper s0), String (StringWrapper s1))) ->
    makeString (s0 ++ s1)

  (ConcatStrings, List vs) ->
    makeString $ concatMap unStringWrapper vs

  (AppendTexts, Tuple (Text t0, Text t1)) ->
    Text $ T.append t0 t1

  (ConcatTexts, List vs) ->
    Text $ T.concat $ map fromValue vs

  (JoinPaths, Tuple (String (StringWrapper s0), String (StringWrapper s1))) ->
    makeString (s0 ++ "/" ++ s1)

  (FileComponents, String (StringWrapper s)) ->
    let (base, ext) = fileComponents s
    in Tuple (makeString base, makeString ext)

  (LowerString, String (StringWrapper s)) ->
    makeString $ map toLower s

  (ElemOf, Tuple (y, List ys)) ->
    Bool (y `elem` ys)

  (MakeTemplate, Tuple (Text t, Text c)) ->
    let (beforeContent, after) = T.breakOn c t
        afterContent = T.drop (T.length c) after
    in Template $ TemplateParts beforeContent afterContent

  (ApplyTemplate, Tuple (Template (TemplateParts beforeContent afterContent), Text t)) ->
    Text $ T.concat [beforeContent, t, afterContent]

  (ToText, String (StringWrapper s)) ->
    Text $ T.pack s

  (ConvertImage, Tuple (Image (ImageWrapper image),
                        ImageConversionSettings (ResizeToWidth widthResized))) ->
    let sizeFactor :: Double
        sizeFactor = fromIntegral (CP.imageWidth image) / fromIntegral widthResized
        heightResized = round (fromIntegral (CP.imageHeight image) / sizeFactor)
    in makeImage $ CPS.resize CPS.defaultOptions widthResized heightResized image

  (RunPandoc, Text contents) ->
    let settings = P.def { P.readerExtensions = P.extensionsFromList [ P.Ext_raw_html ] }
        m = P.writeHtml5 P.def =<< P.readMarkdown settings contents
    in case P.runPure m of
         Left e -> error ("unexpected pandoc error: " ++ show e)
         Right html -> Text $ TL.toStrict $ B.renderHtml html

  _ ->
    error "unexpected combination of function and argument type"
