module FunctionIO
  ( listDirectory
  , isDirectory
  , readTextFile
  , openImage
  , saveImage
  , saveTextFile
  , copyFile
  , makeDir
  ) where

import Prelude hiding (String, FilePath)

import Types.FunctionIO
import Types.Values
import Types.Token (Token(..))
import DependencyGenerator (DepGenM, TokenableTo(..), toTupleToken, runFunctionIO, runFunctionIO_)

import Data.Text (Text)
import qualified Data.Text.IO as T
import qualified Codec.Picture as CP
import qualified System.Directory as SD
import Language.Haskell.TH.Syntax (Lift)


data ListDirectory = ListDirectory deriving (Show, Lift)
instance IsFunctionIO ListDirectory FilePath [FilePath] where
  evalFunctionIO ListDirectory (StringWrapper s) =
    map StringWrapper <$> SD.listDirectory s
  functionIOTouchesFilesystem ListDirectory = False

listDirectory :: TokenableTo FilePath a => a -> DepGenM (Token [FilePath])
listDirectory a = runFunctionIO ListDirectory =<< toToken a


data IsDirectory = IsDirectory deriving (Show, Lift)
instance IsFunctionIO IsDirectory FilePath Bool where
  evalFunctionIO IsDirectory (StringWrapper s) =
    SD.doesDirectoryExist s
  functionIOTouchesFilesystem IsDirectory = False

isDirectory :: TokenableTo FilePath a => a -> DepGenM (Token Bool)
isDirectory a = runFunctionIO IsDirectory =<< toToken a


data ReadTextFile = ReadTextFile deriving (Show, Lift)
instance IsFunctionIO ReadTextFile FilePath Text where
  evalFunctionIO ReadTextFile (StringWrapper s) =
    T.readFile s
  functionIOTouchesFilesystem ReadTextFile = False

readTextFile :: TokenableTo FilePath a => a -> DepGenM (Token Text)
readTextFile a = runFunctionIO ReadTextFile =<< toToken a


data OpenImage = OpenImage deriving (Show, Lift)
instance IsFunctionIO OpenImage FilePath Image where
  evalFunctionIO OpenImage (StringWrapper s) = do
    imageOrig <- CP.readImage s
    case imageOrig of
      Left e -> error ("unexpected error: " ++ e)
      Right image -> pure $ ImageWrapper $ CP.convertRGB8 image
  functionIOTouchesFilesystem OpenImage = False

openImage :: TokenableTo FilePath a => a -> DepGenM (Token Image)
openImage a = runFunctionIO OpenImage =<< toToken a


data SaveImage = SaveImage deriving (Show, Lift)
instance IsFunctionIO SaveImage (Image, FilePath) () where
  evalFunctionIO SaveImage (ImageWrapper image, StringWrapper s) =
    CP.saveJpgImage 90 s $ CP.ImageRGB8 image
  functionIOTouchesFilesystem SaveImage = True

saveImage :: (TokenableTo Image a, TokenableTo FilePath b) => a -> b -> DepGenM ()
saveImage a b = runFunctionIO_ SaveImage =<< toTupleToken a b


data SaveTextFile = SaveTextFile deriving (Show, Lift)
instance IsFunctionIO SaveTextFile (Text, FilePath) () where
  evalFunctionIO SaveTextFile (t, StringWrapper s) =
    T.writeFile s t
  functionIOTouchesFilesystem SaveTextFile = True

saveTextFile :: (TokenableTo Text a, TokenableTo FilePath b) => a -> b -> DepGenM ()
saveTextFile a b = runFunctionIO_ SaveTextFile =<< toTupleToken a b


data CopyFile = CopyFile deriving (Show, Lift)
instance IsFunctionIO CopyFile (FilePath, FilePath) () where
  evalFunctionIO CopyFile (StringWrapper source, StringWrapper target) =
    SD.copyFile source target
  functionIOTouchesFilesystem CopyFile = True

copyFile :: (TokenableTo FilePath a, TokenableTo FilePath b) => a -> b -> DepGenM ()
copyFile a b = runFunctionIO_ CopyFile =<< toTupleToken a b


data MakeDir = MakeDir deriving (Show, Lift)
instance IsFunctionIO MakeDir FilePath () where
  evalFunctionIO MakeDir (StringWrapper s) =
    SD.createDirectory s
  functionIOTouchesFilesystem MakeDir = True

makeDir :: TokenableTo FilePath a => a -> DepGenM ()
makeDir a = runFunctionIO_ MakeDir =<< toToken a
