module Types.Value
  ( Value(..)
  , Valuable(..)
  , makeString
  , makeImage
  ) where

import Prelude hiding (String)
import qualified Prelude

import Types.Values

import Data.Text (Text)
import qualified Codec.Picture as CP
import Language.Haskell.TH.Syntax (Lift)

-- Note: We use a wrapper for the String type in order to be able to define the
-- general Valuable [a] instance further down.  Otherwise it would conflict with
-- our Valuable String instance, since the non-wrapped String type is just an
-- alias for [Char].
--
-- Also note that the Image type does not actually implement Lift, so you
-- shouldn't use it with Inject.  The better approach would be to split the
-- Value type into a compile-time version and a runtime version.
data Value = String String
           | Text Text
           | Bool Bool
           | Image Image
           | ImageConversionSettings ImageConversionSettings
           | Template Template
           | Empty
           | Tuple (Value, Value)
           | List [Value]
  deriving (Eq, Show, Lift)

makeString :: Prelude.String -> Value
makeString = String . StringWrapper

makeImage :: CP.Image CP.PixelRGB8 -> Value
makeImage = Image . ImageWrapper

class Valuable a where
  toValue :: a -> Value
  fromValue :: Value -> a

instance Valuable String where
  toValue = String
  fromValue = \case
    String a -> a
    _ -> error "unexpected"

instance Valuable Text where
  toValue = Text
  fromValue = \case
    Text a -> a
    _ -> error "unexpected"

instance Valuable Bool where
  toValue = Bool
  fromValue = \case
    Bool a -> a
    _ -> error "unexpected"

instance Valuable Image where
  toValue = Image
  fromValue = \case
    Image a -> a
    _ -> error "unexpected"

instance Valuable ImageConversionSettings where
  toValue = ImageConversionSettings
  fromValue = \case
    ImageConversionSettings a -> a
    _ -> error "unexpected"

instance Valuable Template where
  toValue = Template
  fromValue = \case
    Template a -> a
    _ -> error "unexpected"

instance Valuable () where
  toValue () = Empty
  fromValue = \case
    Empty -> ()
    _ -> error "unexpected"

instance (Valuable a, Valuable b) => Valuable (a, b) where
  toValue (a, b) = Tuple (toValue a, toValue b)
  fromValue = \case
    Tuple (va, vb) -> (fromValue va, fromValue vb)
    _ -> error "unexpected"

instance Valuable a => Valuable [a] where
  toValue = List . map toValue
  fromValue = \case
    List a -> map fromValue a
    _ -> error "unexpected"
